/**********************************************************************
  Filename    : Camera Remote Bluetooth IoT
  Description : Control android app Camera Remote Bluetooth remotely.
                Change camera modes: photo or video.
                Change flash modes: torch or auto
                Record video.
                Take photo.
  Version     : ESP32 WROVER and for TFT ST7789 240x240 Color Display            
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/10/30
**********************************************************************/
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"

#include "BluetoothSerial.h"
#include "string.h"

#include "esp_timer.h"
#include "img_converters.h"
#include "fb_gfx.h"
#include "driver/ledc.h"
#include "sdkconfig.h"
#include "esp_camera.h"

#if defined(ARDUINO_ARCH_ESP32) && defined(CONFIG_ARDUHAL_ESP_LOG)
#include "esp32-hal-log.h"
#define TAG ""
#else
#include "esp_log.h"
static const char *TAG = "camera_httpd";
#endif

#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
int led_duty = 0;
bool isStreaming = false;
#ifdef CONFIG_LED_LEDC_LOW_SPEED_MODE
#define CONFIG_LED_LEDC_SPEED_MODE LEDC_LOW_SPEED_MODE
#else
#define CONFIG_LED_LEDC_SPEED_MODE LEDC_HIGH_SPEED_MODE
#endif
#endif

#define CAMERA_MODEL_WROVER_KIT // Has PSRAM
#include "camera_pins.h"


typedef struct
{
    size_t size;  //number of values used for filtering
    size_t index; //current value index
    size_t count; //value count
    int sum;
    int *values; //array to be filled with values
} ra_filter_t;


#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
void enable_led(bool en)
{ // Turn LED On or Off
    int duty = en ? led_duty : 0;
    if (en && isStreaming && (led_duty > CONFIG_LED_MAX_INTENSITY))
    {
        duty = CONFIG_LED_MAX_INTENSITY;
    }
    ledc_set_duty(CONFIG_LED_LEDC_SPEED_MODE, CONFIG_LED_LEDC_CHANNEL, duty);
    ledc_update_duty(CONFIG_LED_LEDC_SPEED_MODE, CONFIG_LED_LEDC_CHANNEL);
    ESP_LOGI(TAG, "Set LED intensity to %d", duty);
}
#endif


#define MCU_MODE 2    //1: ESP32 WROOM 32, 2: ESP32 WROVER
#define DISPLAY_COLOR_MODE 2      // 0: no display, 1: Monochrome (SD1306, black and white), 2: Color (ST7789)

// ESP32 WROOM 32: I2C Connection *************************************
// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
// The pins for I2C are defined by the Wire-library. 
// On an arduino UNO:       A4(SDA), A5(SCL)
// On an arduino MEGA 2560: 20(SDA), 21(SCL)
// On an arduino LEONARDO:   2(SDA),  3(SCL), ...

// On ESP32 : 21(SDA), 22(SCL) ****************************************
// I2C:
// Default hardware pins: SDA (21), SCL (22)
//#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
//#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32

// uncomment and replace with statement bellow for this display
//Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// SD1306 *************************************************************
//#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32
//#define OLED_MOSI   23
//#define OLED_CLK   18
//#define OLED_DC    15     //19
//#define OLED_CS    5
//#define OLED_RESET 4
//      
// uncomment and replace with statement bellow for this display   
// Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, OLED_MOSI, OLED_CLK, OLED_DC, OLED_RESET, OLED_CS);

// ESP32 WROVER: SPI Connection *****************************************
// SPI: VSPI (software spi: serial peripheral interface)
// Default hardware pins: MOSI (23), CLK (18), CS (5)
// Ueer selective pins: DC, RESET
// ST7789 ***************************************************************
#define TFT_CS         -1      //10
#define TFT_RST        4      //9 // Or set to -1 and connect to Arduino RESET pin
#define TFT_DC         15     //8
//#define TFT_MOSI 23     //11  // Data out
//#define TFT_SCLK 18     //13  // Clock out
// For 1.14", 1.3", 1.54", 1.69", and 2.0" TFT with ST7789:
// Hardware SPI
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);
// Software SPI
// Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);
// replace this statement with other display for correct display and versions
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);

// Bluetooth Status
int CONNECTED = 0;

BluetoothSerial SerialBT;

void setup() {

  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);    //disable brownout detector   ds

//  SerialBT.begin("Camera Remote Bluetooth");    //Bluetooth device name
  SerialBT.begin("Bluetooth Camera", false);    //Bluetooth device name, slave mode (connect from android device) 

  Serial.begin(115200);

  SerialBT.register_callback(callback);
  
  Serial.println("\nThe camera started, now you can connect (paired) from android remote mode by bluetooth!");
  
  delay(2000);
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 1);    //enable brownout detector

  if (CONNECTED) {
    // no codes will be run here
  }



  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;
  
  // if PSRAM IC present, init with UXGA resolution and higher JPEG quality
  //                      for larger pre-allocated frame buffer.
  if(psramFound()){
    config.frame_size = FRAMESIZE_UXGA;
    config.jpeg_quality = 10;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_SVGA;
    config.jpeg_quality = 12;
    config.fb_count = 1;
  }

//  config.frame_size = FRAMESIZE_QVGA;
//  config.jpeg_quality = 10;
//  config.fb_count = 1;

#if defined(CAMERA_MODEL_ESP_EYE)
  pinMode(13, INPUT_PULLUP);
  pinMode(14, INPUT_PULLUP);
#endif

  // camera init
  esp_err_t err = esp_camera_init(&config);
  if (err != ESP_OK) {
    Serial.printf("Camera init failed with error 0x%x", err);
    return;
  } else {
    Serial.println("Camera initialized successfully.");
  }

  sensor_t * s = esp_camera_sensor_get();
  // initial sensors are flipped vertically and colors are a bit saturated
  if (s->id.PID == OV3660_PID) {
    s->set_vflip(s, 1); // flip it back
    s->set_brightness(s, 1); // up the brightness just a bit
    s->set_saturation(s, -2); // lower the saturation
  }
  // drop down frame size for higher initial frame rate
  s->set_framesize(s, FRAMESIZE_QVGA);

#if defined(CAMERA_MODEL_M5STACK_WIDE) || defined(CAMERA_MODEL_M5STACK_ESP32CAM)
  s->set_vflip(s, 1);
  s->set_hmirror(s, 1);
#endif



  
}

void callback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param) {
  if(event == ESP_SPP_SRV_OPEN_EVT){
    Serial.println("Client Connected");
    //digitalWrite(LED,HIGH);
    CONNECTED = true;
  }
 
  if(event == ESP_SPP_CLOSE_EVT ){
    Serial.println("Client disconnected");
    //digitalWrite(LED,LOW);
    CONNECTED = false;
  }
}

String dataValues[15];  
// clears the data value array
void clearDataValues() {
  int numberOfValues = 10;    //(int)sizeof(dataValues)/sizeof("hello");
  for (int i = 0; i < numberOfValues; i++) {
    dataValues[i] = "";
  }
}
void parseData(String dataString) {
  clearDataValues();

  // data delimeter: ":"
  char delimeter = ':';
  int findStart = 0;
  int findIndex = 0;
  int findCount = 0;

  findIndex = dataString.indexOf(delimeter, findStart);
  if (findIndex > 0) {
    dataValues[findCount] = dataString.substring(findStart, findIndex);
    findCount ++;
    findStart = findIndex + 1;

    while (findIndex > 0) {
      findIndex = dataString.indexOf(delimeter, findStart);
      if (findIndex > 0) {
        dataValues[findCount] = dataString.substring(findStart, findIndex);
        findCount ++;
        findStart = findIndex + 1;        
      }
    } 
  }
}

bool imageDataProcessing = false;

void processData(String dataString) {

  //dataString = datastring;
  if (!dataString.startsWith("imagepixels")) {
    Serial.println("processData: " + dataString);
  }

  // parse data string into values
  parseData(dataString);

  // process parsed data values
  // data name: dataValues[0], 'photo', 'video', 'iot', 'imagedata', 'imagefixels', 'received', , 'imagefixels_color'
//
//  String dataName = dataValues[0];
  

  dataString = "";
}

void processImageData(String header) {
  // image data header received: "imagedata:xxxxx:16bit" image data pixel count, and color mode (16 bit or 8 bit)
//
//  if (!header.startsWith("imagedata")) {
//    imageDataProcessing = false;
//    return;
//  }

  Serial.println(header);

  parseData(header);
}


void processAction() {


        if (CONNECTED) {

//          SerialBT.println(actionString);
//          SerialBT.write(13);
//          SerialBT.write(10);
      
//          ACTION_STATUS = 1;
    
//          Serial.println(actionString);          
        }

        delay(100);     

//
//    if (ACTION_STATUS == 1) {
//
////      if (photo && video && flash && action && camera && displayMode) {
////        ACTION_STATUS = 0;
////        delay(10);      
////      }
//    
//    }

}

uint8_t* frameData;

esp_err_t camera_capture(){
    camera_fb_t * fb = NULL;
    size_t _jpg_buf_len = 0;
    uint8_t *_jpg_buf = NULL;
  
    //capture a frame
    fb = esp_camera_fb_get();
    if (!fb) {
        //ESP_LOGE(TAG, "Frame buffer could not be acquired");
        Serial.println("Frame buffer could not be acquired");
        delay(200);
        return ESP_FAIL;
    }

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;

    //replace this with your own function
    //display_image(fb->width, fb->height, fb->pixformat, fb->buf, fb->len);

    //return the frame buffer back to be reused
    if (fb)
    {
        Serial.print("width: ");
        Serial.print(fb->width);
        Serial.print(", height: ");
        Serial.print(fb->height);
        Serial.print(", len: ");
        Serial.println(_jpg_buf_len);

        if (CONNECTED) {
          
          // send frame data header
          SerialBT.print("frame_data");
          SerialBT.print(":width:");
          SerialBT.print(fb->width);
          SerialBT.print(":height:");
          SerialBT.print(fb->height);
          SerialBT.print(":len:");
          SerialBT.println(_jpg_buf_len);

          // send frame data
          //frameData = frameData[_jpg_buf_len];
          //memcpy(frameData, _jpg_buf, _jpg_buf_len);  
          
          //frameData = (uint8_t*)malloc(_jpg_buf_len * sizeof(uint8_t));
          //memmove(frameData, _jpg_buf, _jpg_buf_len);  
          
          //delay(10);       
          //SerialBT.write(frameData, _jpg_buf_len);         
          SerialBT.write(_jpg_buf, _jpg_buf_len);         

          // send new line
          SerialBT.write(13);
          SerialBT.write(10);
        }
        //delay(10);
        esp_camera_fb_return(fb);
        fb = NULL;
        _jpg_buf = NULL;

        
    } else if (_jpg_buf)
    {
        free(_jpg_buf);
        _jpg_buf = NULL;
    }    


    return ESP_OK;
}

void changeFrameSize(String dataString) {

  parseData(dataString);

  String frameSize = dataValues[1];
  framesize_t fs = FRAMESIZE_QVGA;
  
  if (frameSize.equalsIgnoreCase("small")) {
    fs = FRAMESIZE_QQVGA;  //FRAMESIZE_QQVGA;
  } else if (frameSize.equalsIgnoreCase("medium")) {
    fs = FRAMESIZE_QVGA;  //FRAMESIZE_QVGA;
  } else if (frameSize.equalsIgnoreCase("large")) {
    fs = FRAMESIZE_HVGA;  //FRAMESIZE_HVGA;  //FRAMESIZE_VGA;
  }

  sensor_t *s = esp_camera_sensor_get();
  int res = 0;
  res = s->set_framesize(s, fs);

}

bool takingPicture = false;
void takePictureAndSend() {

  sensor_t *s = esp_camera_sensor_get();
  //framesize_t fs = s->status.framesize;
  int res = 0;
  res = s->set_framesize(s, FRAMESIZE_HD);

  delay(500);

  camera_fb_t * fb = NULL;
  size_t _jpg_buf_len = 0;
  uint8_t *_jpg_buf = NULL;

  // wait for the auto exposure sets
  for (int i = 0; i < 6; i++) {
    fb = esp_camera_fb_get();
    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
    delay(100);  
  }

  
  fb = esp_camera_fb_get();
  if (!fb) {
    //ESP_LOGE(TAG, "Frame buffer could not be acquired");
    Serial.println("Frame buffer could not be acquired");
      
    //send failed status message
    // send frame data header
    SerialBT.print("picture_data");
    SerialBT.print(":width:0");
    SerialBT.print(":height:0");
    SerialBT.print(":len:0:");
     // send new line
    SerialBT.write(13);
    SerialBT.write(10);  
     
  } else {

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    
    
    // send frame data header
    SerialBT.print("picture_data");
    SerialBT.print(":width:");
    SerialBT.print(fb->width);
    SerialBT.print(":height:");
    SerialBT.print(fb->height);
    SerialBT.print(":len:");
    SerialBT.println(_jpg_buf_len);
         
    SerialBT.write(_jpg_buf, _jpg_buf_len);   
    // send new line
    SerialBT.write(13);
    SerialBT.write(10);  

    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
  }


  delay(100);

//  sensor_t *s2 = esp_camera_sensor_get();  
//  s2 = esp_camera_sensor_get();
//  res = 0;
//  res = s2->set_framesize(s2, fs);
//
//  delay(100);
  
  takingPicture = false;
}


void loop() {

//  currentTime = millis();
//  if (modeChangedTime < 0) {
//    modeChangedTime = currentTime;
//  }

  if (CONNECTED) {


    while(SerialBT.available())
    {
      
      String received = SerialBT.readStringUntil('\n');
      Serial.println(received);
      if (!takingPicture) {
        if (received.startsWith("framesize")) {
          changeFrameSize(received);
        } else if (received.startsWith("frameget")) {
          takingPicture = true;
          delay(50);
          takePictureAndSend();      
        }      
      } else {
        delay(100);
      }

    }

    if (!takingPicture) {
      esp_err_t result = camera_capture();
    }
    
    processAction();

  } else { //end of: if (CONNECTED)

    // not connected
//    digitalWrite(PIN_LED_PHOTO, LOW);
//    digitalWrite(PIN_LED_VIDEO, LOW);

   
  }

  delay(50);
}
